#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <utils.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <dirent.h>

//#include <cutils/log.h>

#include "libsensorhub.h"
#include "bist.h"
#include "message.h"

#define SNLEN 16
#define KEYLEN 128
#define VALLEN 128
#define DESLEN 256

#define UNIT_RAW 0
#define UNIT_STD 1


enum sc_config_index
{
	INDEX_PEDOMETER_MODE,
	INDEX_PEDOMETER_N,
	INDEX_ACT_MODE,
	INDEX_ACT_CLSMASK,
	INDEX_ACT_N,
	INDEX_ACT_DUTY_M,
	INDEX_GFLICK_CLSMASK,
	INDEX_GFLICK_LEVEL,
	INDEX_SHAKING_SENSITIVITY,
	INDEX_STAP_CLSMASK,
	INDEX_STAP_LEVEL,
	INDEX_DTWGSM_LEVEL,
	INDEX_DTWGSM_DST,
	INDEX_DTWGSM_TEMPLATE,
	INDEX_WPD_SCOUNTERMODE,
	INDEX_WPD_NSC,
	INDEX_INSTANTACTIVITY_CLSMASK,
	INDEX_EARTOUCH_CLSMASK,
	INDEX_PDR_USER_H,	/* user's height */
	INDEX_PDR_FLOOR_H,	/* building's floor height */
	INDEX_PDR_INIT_X,	/* init position x (East) */
	INDEX_PDR_INIT_Y,	/* init position y (North) */
	INDEX_PDR_INIT_FL,	/* init floor level */
	INDEX_PDR_N,		/* report interval */
	INDEX_PDR_6DOF,	/* use 6DOF as rmat input, or not */
	INDEX_LIFT_MASK,
	INDEX_COUNT
};


typedef struct
{
	int id;
	int size;
	char* val;
} sc_config_t;

#define MAX_CONFIG 256
int config_count = 0;
sc_config_t configs[MAX_CONFIG];

static int ispropset = 0;
static int isstarting = 0;
static int rate_config = 1;
static int delay_config = 0;
static int cont_config = 0;
#define FMT_BYTE  1
#define FMT_WORD  2
#define FMT_DWORD 4
static int data_format = FMT_BYTE;
static int sample_count = 0;

struct sc_s;

typedef int (*fn_proc)(struct sc_s *sc);
typedef int (*fn_start)(struct sc_s *sc, handle_t handle);
typedef int (*fn_stop)(struct sc_s *sc, handle_t handle);
typedef int (*fn_config)(struct sc_s *__sc, handle_t handle);
typedef void (*fn_dump)(struct sc_s *sc, int fd, char *buf, int size);

typedef struct sc_s
{
	char name[16];
	char *desc;
	int pack_size;
	fn_proc proc;
	fn_start start;
	fn_stop stop;
	fn_config config;
	fn_dump dump;
	enum sc_config_index props[16];
} sc_t;

static char* ts2str(int64_t ts)
{
	static char buf[256];
	struct tm *t = gmtime((time_t*)&ts);
	snprintf(buf, 256, "%lld", ts);
	return buf;
}

static int general_proc(sc_t *sc)
{
	char buf[1024];
	handle_t handle = NULL;
	int fd = -1;
	handle = psh_open_session_with_name(sc->name);
	if (handle == NULL) {
		fprintf(stderr, "Failed open session. Sensor name: %s\n", sc->name);
		return -1;
	}
	if (sc->config(sc, handle) < 0) {
		return -1;
	}
	if (sc->start(sc, handle) < 0) {
		return -1;
	}
	fd = psh_get_fd(handle);
	int count = 0;
	while (sample_count <= 0 || count < sample_count) {
		sc->dump(sc, fd, buf, sizeof(buf));
		fprintf(stdout, "%s", buf);
		++count;
	}
	sc->stop(sc, handle);
	psh_close_session(handle);
	return 0;
}

static int general_config(sc_t *sc, handle_t handle)
{
	int i = 0;
	for (i = 0; i < config_count; ++i) {
		sc_config_t *pcfg = configs +i;
		if (psh_set_property_with_size(handle, pcfg->id, pcfg->size, pcfg->val) < 0) {
			fprintf(stderr, "Failed to set property. Sensor name: %s, Prop id: %d, Size: %d\n", sc->name, pcfg->id, pcfg->size);
		}
	}
	return 0;
}


static int dtwgs_config(sc_t *sc, handle_t handle)
{
	fprintf(stderr, "currently DTWGS property setting is not supported!\n");
	return 0;
}

static int general_start(sc_t *sc, handle_t handle)
{
	int ret = 0;
	ret = psh_start_streaming_with_flag(handle, rate_config, delay_config, cont_config);
	if (ret < 0) {
		fprintf(stderr, "Failed to start streaming. Sensor: %s\n", sc->name);
	}
	return ret;
}

static int general_stop(sc_t *sc, handle_t handle)
{
	psh_stop_streaming(handle);
	return 0;
}

static void general_dump(sc_t *sc, int fd, char *buf, int size)
{
	char fbuf[2048];
	int fsize = 0;
	int finx = 0;
	int64_t ts = 0;
	int buf_write = 0;

	if (sc->pack_size > 0) {
		fsize = read(fd, fbuf, sc->pack_size);
	} else {
		fsize = read(fd, fbuf, sizeof(fbuf));
	}
	finx = 0;

	ts = *(int64_t*)fbuf;
	finx += sizeof(int64_t);
	buf_write = snprintf(buf, size, "%s ", ts2str(ts));
	buf += buf_write;
	size -= buf_write;

	while (finx < fsize) {
		int d = 0;
		switch (data_format) {
		case FMT_BYTE:
			d = *(char*)(fbuf + finx);
			finx += sizeof(char);
			buf_write = snprintf(buf, size, "%d ", d);
			buf += buf_write;
			size -= buf_write;
			break;
		case FMT_WORD:
			d = *(short*)(fbuf + finx);
			finx += sizeof(short);
			buf_write = snprintf(buf, size, "%d ", d);
			buf += buf_write;
			size -= buf_write;
			break;
		case FMT_DWORD:
			d = *(int*)(fbuf + finx);
			finx += sizeof(int);
			buf_write = snprintf(buf, size, "%d ", d);
			buf += buf_write;
			size -= buf_write;
			break;
		}
	}
	snprintf(buf, size, "\n");
}

static void dummy_dump(sc_t *sc, int fd, char *buf, int size)
{
	char t[1024];
	read(fd, t, sizeof(t));
}

static sc_t custom_sc = {
		"",
		"custom",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
};
static sc_t scs[] = {
	{
		"ACCEL",
		"accelerometer",
		sizeof(struct accel_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"GYRO",
		"gyroscope",
		sizeof(struct gyro_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"COMPS",
		"compass",
		sizeof(struct compass_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"BARO",
		"barometer",
		sizeof(struct baro_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"ALS_P",
		"ambient light sensor",
		sizeof(struct als_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"PS_P",
		"proximity",
		sizeof(struct ps_phy_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"TERMC",
		"terminal context",
		sizeof(struct tc_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"LPE_P",
		"LPE",
		sizeof(struct lpe_phy_data),
		general_proc, //TODO
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"ACC1",
		"secondary accelerometer",
		sizeof(struct accel_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"GYRO1",
		"secondary gyroscope",
		sizeof(struct gyro_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"COMP1",
		"secondary compass",
		sizeof(struct compass_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"ALS1",
		"secondary ambient light sensor",
		sizeof(struct als_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"PS1",
		"secondary proximity",
		sizeof(struct ps_phy_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"BARO1",
		"secondary barometer",
		sizeof(struct baro_raw_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"PHYAC",
		"physical activity classifier",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_ACT_MODE,
			INDEX_ACT_CLSMASK,
			INDEX_ACT_N,
			INDEX_ACT_DUTY_M,
			-1
		}
	},
	{
		"GSSPT",
		"gesture spotting collecter",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"GSFLK",
		"gesture flick detecter",
		sizeof(struct gesture_flick_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_GFLICK_CLSMASK,
			INDEX_GFLICK_LEVEL,
			-1
		}
	},
	{
		"RVECT",
		"rotation vector",
		sizeof(struct rotation_vector_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"GRAVI",
		"gravity",
		sizeof(struct gravity_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"LACCL",
		"linear accelerometer",
		sizeof(struct linear_accel_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"ORIEN",
		"orientation",
		sizeof(struct orientation_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"COMPC",
		"compass caliburation",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"GYROC",
		"gyroscope caliburation",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"9DOF",
		"9 degree of freedom",
		sizeof(struct ndof_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"MAGHD",
		"magnetic heading",
		sizeof(struct mag_heading_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"SHAKI",
		"shaking detect",
		sizeof(struct shaking_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_SHAKING_SENSITIVITY,
			-1
		}
	},
	{
		"MOVDT",
		"moving detect",
		sizeof(struct md_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"STAP",
		"tapping detect",
		sizeof(struct stap_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_STAP_CLSMASK,
			INDEX_STAP_LEVEL,
			-1
		}
	},
	{
		"PZOOM",
		"pan zoom",
		sizeof(struct pz_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"DVPOS",
		"device position",
		sizeof(struct device_position_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"SCOUN",
		"step counter",
		sizeof(struct stepcounter_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_WPD_SCOUNTERMODE,
			INDEX_WPD_NSC,
			-1
		}
	},
	{
		"SIGMT",
		"significant motion detector",
		sizeof(struct sm_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"6AGRV",
		"6A game rotation vector",
		sizeof(struct game_rotation_vector_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"6AMRV",
		"6A geomatric rotation vector",
		sizeof(struct geomagnetic_rotation_vector_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"6DOFG",
		"6DOF AG",
		sizeof(struct ndofag_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"6DOFM",
		"6DOF AM",
		sizeof(struct ndofam_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"LIFT",
		"lift gesture detector",
		sizeof(struct lift_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_LIFT_MASK,
			-1
		}
	},
	{
		"DTWGS",
		"DTW gesture classifier",
		sizeof(struct dtwgs_data),
		general_proc,
		general_start,
		general_stop,
		dtwgs_config,
		general_dump,
		{
			INDEX_DTWGSM_LEVEL,
			INDEX_DTWGSM_DST,
			INDEX_DTWGSM_TEMPLATE,
			-1
		}
	},
	{
		"GSPX",
		"glyph gesture classifier",
		sizeof(struct gesture_hmm_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			-1
		}
	},
	{
		"GSETH",
		"ear-touch gesture detector",
		sizeof(struct gesture_eartouch_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_EARTOUCH_CLSMASK,
			-1
		}
	},
	{
		"PDR",
		"pedestrain dead-reconing",
		sizeof(struct pdr_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_PDR_USER_H,
			INDEX_PDR_FLOOR_H,
			INDEX_PDR_INIT_X,
			INDEX_PDR_INIT_Y,
			INDEX_PDR_INIT_FL,
			INDEX_PDR_N,
			INDEX_PDR_6DOF,
			-1
		}
	},
	{
		"ISACT",
		"instant activity classifier",
		sizeof(struct instant_activity_data),
		general_proc,
		general_start,
		general_stop,
		general_config,
		general_dump,
		{
			INDEX_INSTANTACTIVITY_CLSMASK,
			-1
		}
	},
	{
		"BIST",
		"BIST",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		dummy_dump,
		{
			-1
		}
	},
	{
		"EVENT",
		"EVENT",
		0,
		general_proc,
		general_start,
		general_stop,
		general_config,
		dummy_dump,
		{
			-1
		}
	},
	{ "", NULL, 0, NULL, NULL, NULL, NULL, NULL, { -1 } }
};

static void usage(int argc, char *argv[])
{
	printf("Usage:\n");
	printf("  %s -l\n", argv[0]);
	printf("  %s [-i<T>] [-b<S>] [-p<P>] [-f<F>] [-c<C>] -r <SENSOR-NAME>\n", argv[0]);
	printf("  %s -h\n", argv[0]);
	printf("OPTIONS:\n");
	printf("-l  --list           List all available SENSOR-NAMEs\n");
	printf("-i  --interval=T     Set the interval time between each sample.\n");
	printf("                     \"T\" is in milliseconds.\n");
	printf("                     Default is 1000\n");
	printf("-b  --standby=S      Set how sensor behave when system sleeps.\n");
	printf("                     \"S\" may be one of the following values:\n");
	printf("                       wake: when system sleeps, sensor continues working\n");
	printf("                             and wakes up system whenever it reports new data\n");
	printf("                       continue: when system sleeps, sensor continues working\n");
	printf("                             but doesn't report even on new data coming, so system wouldn't be waken.\n");
	printf("                       stop: when system sleeps, sensor stops working\n");
	printf("                     Default is \"stop\"\n");
	printf("-p  --property=P     Set sensor property.\n");
	printf("                     Format is: \"<id>,<byte 1>,<byte 2>,...<byte n>\"\n");
	printf("                     All bytes are in decimal\n");
	//printf("-R  --reset          Reset all configurations and properties\n");
	printf("-r  --run            Start the sensor and display data reports\n");
	printf("-f  --data-format=F  Set the format of sensor data report\n");
	printf("                     \"F\" may be one of the following values:\n");
	printf("                       \"%%B\": show data in bytes\n");
	printf("                       \"%%W\": show data in words(2 bytes integer)\n");
	printf("                       \"%%D\": show data in dwords(4 bytes integer)\n");
	printf("                     All are in decimal\n");
	printf("                     Default is %%B\n");
	printf("-c  --sample-count=C Set how much samples to be displayed before sensor stopped.\n");
	printf("                     If \"C\" is less then or equal to 0, than it will never stop automatically;\n");
	printf("                     otherwise it will report exactly \"C\" samples and then stop\n");
	printf("                     Default is 0\n");
	printf("-h  --help           Show this message\n");
}


static sc_t* find(const char *sn)
{
	sc_t *p = scs;
	while (p->name[0] != 0) {
		if (strncmp(sn, p->name, sizeof(p->name)) == 0) {
			return p;
		}
		++p;
	}
	return NULL;
}

static void list()
{
	const char * path_template = "/sys/class/hwmon/hwmon%d/device";
	int hwmon_num;
	char path[256];
	char path_file[256];
	int found = 0;
	char content[1024];
	for (hwmon_num = 0; hwmon_num < 10; ++hwmon_num) {
		snprintf(path, sizeof(path), path_template, hwmon_num);
		DIR *dir = opendir(path);
		if (dir == NULL) {
			continue;
		}
		struct dirent *ent = readdir(dir);
		while (ent != NULL) {
			if (strncmp(ent->d_name, "fw_version", sizeof(ent->d_name)) == 0) {
				found = 1;
				break;
			}
			ent = readdir(dir);
		}
		closedir(dir);
		if (found) {
			break;
		}
	}
	if (!found) {
		fprintf(stderr, "/sys/class/hwmon/hwmon<n>/device/fw_version not found\n");
		return;
	}

	FILE *f = NULL;

	strncpy(path_file, path, sizeof(path_file));
	strncat(path_file, "/trace", sizeof(path_file));

	f = fopen(path_file, "rt");
	fflush(f);
	fclose(f);

	strncpy(path_file, path, sizeof(path_file));
	strncat(path_file, "/control", sizeof(path_file));

	f = fopen(path_file, "wt");
	fprintf(f, "0 11 0 -1 -1 -1 -1\n");
	fclose(f);

	strncpy(path_file, path, sizeof(path_file));
	strncat(path_file, "/trace", sizeof(path_file));

	f = fopen(path_file, "rt");
	char *line = fgets(content, sizeof(content), f);

	/*prevent redundent id*/
	int id_set[100];
	int id_inx = 0;

	while (line != NULL) {
		const char *start = "***** Sensor";
		char sensor_name[32];
		int sensor_id = -1;
		char sensor_desc[64] = "";
		if (strstr(line, start) == line) {
			char * p = content + strlen(start);
			while (*p == ' ') {
				++p;
			}
			char * q = p;
			while (*q != '(') {
				++q;
			}
			*q = 0;
			strncpy(sensor_name, p, sizeof(sensor_name));
			++q;
			sensor_id = atoi(q);
			sc_t *s = find(sensor_name);
			if (s != NULL) {
				strncpy(sensor_desc, s->desc, sizeof(sensor_desc));
			}
			int j;
			for (j = 0; j < id_inx; ++j) {
				if (id_set[j] == sensor_id) {
					break;
				}
			}
			if (j == id_inx) {
				fprintf(stdout, "Name:%s,Id:%d,Desc:%s\n", sensor_name, sensor_id, sensor_desc);
				if (id_inx < sizeof(id_set)/sizeof(id_set[0])) {
					id_set[id_inx] = sensor_id;
					++id_inx;
				}
			}
		}
		line = fgets(content, sizeof(content), f);
	}
	fclose(f);
}

static void set_interval(char *arg)
{
	rate_config = 1000/atoi(arg);
}

static void set_standby(char *arg)
{
	if (strncmp(arg, "wake", 16) == 0) {
		cont_config = 1;
	} else if (strncmp(arg, "continue", 16) == 0) {
		cont_config = 2;
	} else if (strncmp(arg, "stop", 16) == 0) {
		cont_config = 0;
	}
}

static int set_property(char *arg)
{
	char _arg[1024];
	strncpy(_arg, arg, sizeof(_arg));
	sc_config_t cfg;
	char *tok = strtok(_arg, ",");
	int count = 0;
	while (tok != NULL) {
		++count;
		tok = strtok(NULL, ",");
	}
	if (count <= 0)
		return -1;
	int id = -1;
	char *val = (char*)malloc(count - 1);
	int i;
	strncpy(_arg, arg, sizeof(_arg));
	tok = strtok(_arg, ",");
	for (i = 0; i < count; ++i) {
		if (i == 0) {
			id = atoi(tok);
		} else {
			val[i - 1] = atoi(tok);
		}
		tok = strtok(NULL, ",");
	}
	cfg.id = id;
	cfg.size = count - 1;
	cfg.val = val;
	for (i = 0; i < config_count; ++i) {
		if (configs[i].id == cfg.id) {
			free(configs[i].val);
			configs[i] = cfg;
			return 0;
		}
	}
	if (i < MAX_CONFIG) {
		configs[i] = cfg;
		++config_count;
		return 0;
	} else {
		free(val);
		return -1;
	}
}

static int set_data_format(char *arg)
{
	char *tok = strchr(arg, '%');
	if (tok == NULL) {
		return -1;
	}
	switch (*(tok + 1)) {
	case 'B':
		data_format = FMT_BYTE;
		break;
	case 'W':
		data_format = FMT_WORD;
		break;
	case 'D':
		data_format = FMT_DWORD;
		break;
	default:
		return -1;
	}
	return 0;
}

static int store_configs()
{
	int i, j;
	const char *tmpdir = "/tmp";
	const char *tmpfile = "/tmp" "/" "pshcmd-configs";
	mkdir(tmpdir, S_IRWXU);
	FILE * fp = fopen(tmpfile, "wt");
	if (fp == NULL) {
		return -1;
	}
	fprintf(fp, "%d,%d,%d,%d,%d\n", rate_config, delay_config, cont_config, data_format, sample_count);
	for (i = 0; i < config_count; ++i) {
		fprintf(fp, "%d", configs[i].id);
		for (j = 0; j < configs[i].size; ++j) {
			fprintf(fp, ",%d", configs[i].val[j]);
		}
		fprintf(fp, "\n");
	}
	fclose(fp);
	return 0;
}

static int load_configs()
{
	int i;
	char buf[1024];
	const char *tmpdir = "/tmp";
	const char *tmpfile = "/tmp" "/" "pshcmd-configs";
	FILE * fp = fopen(tmpfile, "rt");
	if (fp == NULL) {
		return -1;
	}
	char *line = NULL;
	line = fgets(buf, sizeof(buf), fp);
	if (line == NULL) {
		fclose(fp);
		return -1;
	}
	sscanf(buf, "%d,%d,%d,%d,%d", &rate_config, &delay_config, &cont_config, &data_format, &sample_count);
	line = fgets(buf, sizeof(buf), fp);
	while (line != NULL) {
		set_property(buf);
		line = fgets(buf, sizeof(buf), fp);
	}
	fclose(fp);
	return 0;
}

static int reset_configs()
{
	int i;
	const char *tmpdir = "/tmp";
	const char *tmpfile = "/tmp" "/" "pshcmd-configs";
	FILE * fp = fopen(tmpfile, "wt");
	if (fp == NULL) {
		return 0;
	}
	fclose(fp);
	return 0;
}

int main(int argc, char *argv[])
{
	memset(configs, 0, sizeof(configs));

	static struct option opts[] = {
		{"list", 0, NULL, 'l'},
		{"interval", 1, NULL, 'i'},
		{"standby", 1, NULL, 'b'},
		{"property", 1, NULL, 'p'},
		{"reset", 0, NULL, 'R'},
		{"run", 2, NULL, 'r'},
		{"data-format", 1, NULL, 'f'},
		{"sample-count", 1, NULL, 'c'},
		{"help", 0, NULL, 'h'},
		{0, 0, NULL, 0}
	};
	int index, opt;
	char sn[SNLEN];

	if (argc <= 1) {
		usage(argc, argv);
		exit(EXIT_SUCCESS);
	}

	//load_configs();

	while (1) {

		opt = getopt_long(argc, argv, "li:b:p:Rr::f:c:h", opts, &index);
		if (opt == -1) {
			break;
		}

		switch (opt) {
		case 'l':
			list();
			exit(EXIT_SUCCESS);
		case 'i':
			ispropset = 1;
			set_interval(optarg);
			break;
		case 'b':
			ispropset = 1;
			set_standby(optarg);
			break;
		case 'p':
			ispropset = 1;
			set_property(optarg);
			break;
		case 'R':
			reset_configs();
			exit(EXIT_SUCCESS);
		case 'r':
			strncpy(sn, argv[argc - 1], SNLEN);
			isstarting = 1;
			break;
		case 'f':
			if (set_data_format(optarg) < 0) {
				fprintf(stderr, "Invalid format: %s\n", optarg);
			} else {
				ispropset = 1;
			}
			break;
		case 'c':
			ispropset = 1;
			sample_count = atoi(optarg);
			break;
		case 'h':
			usage(argc, argv);
			exit(EXIT_SUCCESS);
		default:
			usage(argc, argv);
			exit(EXIT_SUCCESS);
		}
	}

	/*if (ispropset) {
		store_configs();
	}*/
	if (isstarting) {
		sc_t *p = find(sn);
		if (p == NULL) {
			p = &custom_sc;
			strncpy(p->name, sn, sizeof(p->name));
		}
		if (p->proc(p) < 0) {
			exit(-1);
		}
	}

	return EXIT_SUCCESS;
}

